"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const node_worker_threads_1 = require("node:worker_threads");
const node_fs_1 = require("node:fs");
const graphql_1 = require("graphql");
const load_1 = require("graphile-config/load");
const graphile_config_1 = require("graphile-config");
const operationPaths_1 = require("./operationPaths");
const rulesContext_1 = require("./rulesContext");
const DepthVisitor_1 = require("./DepthVisitor");
if (node_worker_threads_1.isMainThread) {
    throw new Error("This script is designed to be called by `scan.ts`, but it's the main thread");
}
if (!node_worker_threads_1.parentPort) {
    throw new Error("This script is designed to be called by `scan.ts`, but there's no parent port");
}
const definitelyParentPort = node_worker_threads_1.parentPort;
async function main() {
    const { configPath, overrideConfig } = node_worker_threads_1.workerData;
    const rawConfig = await (0, load_1.loadConfig)(configPath);
    const config = (0, graphile_config_1.resolvePresets)([
        rawConfig ?? {},
        { gqlcheck: overrideConfig },
    ]);
    const { gqlcheck: { schemaSdlPath = `${process.cwd()}/schema.graphql` } = {}, } = config;
    const schemaString = (0, node_fs_1.readFileSync)(schemaSdlPath, "utf8");
    const schema = (0, graphql_1.buildASTSchema)((0, graphql_1.parse)(schemaString));
    {
        const schemaErrors = (0, graphql_1.validateSchema)(schema);
        if (schemaErrors.length) {
            console.error(schemaErrors);
            throw new Error(`GraphQL schema is invalid.`);
        }
    }
    async function checkDocument(req) {
        const { sourceString, sourceName } = req;
        const source = new graphql_1.Source(sourceString, sourceName);
        const document = (0, graphql_1.parse)(source);
        const operationDefinitions = document.definitions.filter((o) => o.kind === graphql_1.Kind.OPERATION_DEFINITION);
        if (operationDefinitions.length === 0) {
            return {
                sourceName,
                operations: [],
                errors: [{ message: "Could not find any operations in this document" }],
            };
        }
        // TODO: regular validation
        const validationErrors = (0, graphql_1.validate)(schema, document);
        if (validationErrors.length > 0) {
            return {
                sourceName,
                operations: [],
                errors: validationErrors.map((e) => e.toJSON?.() ?? (0, graphql_1.formatError)(e)),
            };
        }
        const typeInfo = new operationPaths_1.TypeAndOperationPathInfo(schema);
        const errors = [];
        function onError(error) {
            errors.push(error.toJSON?.() ?? (0, graphql_1.formatError)(error));
        }
        const rulesContext = new rulesContext_1.RulesContext(schema, document, typeInfo, config, onError);
        const visitor = (0, graphql_1.visitInParallel)([(0, DepthVisitor_1.DepthVisitor)(rulesContext)]);
        (0, graphql_1.visit)(document, (0, graphql_1.visitWithTypeInfo)(typeInfo, visitor));
        const operations = operationDefinitions.map((operationDefinition) => {
            const operationName = operationDefinition.name?.value;
            const operationKind = operationDefinition.operation;
            return {
                operationName,
                operationKind,
            };
        });
        return {
            sourceName,
            errors,
            operations,
        };
    }
    definitelyParentPort.on("message", (req) => {
        if (req === "STOP") {
            process.exit(0);
        }
        checkDocument(req).then((result) => {
            definitelyParentPort.postMessage(result);
        }, (e) => {
            console.dir(e);
            process.exit(1);
        });
    });
    definitelyParentPort.postMessage("READY");
}
main().catch((e) => {
    console.error(e);
    process.exit(1);
});
